#include <map>
#include <GL/gl.h>
#include <stdio.h>
#include <iostream>
#include <stdexcept>
#include <sstream>
#include <cmath>

#include "game.h"
#include "rock.h"
#include "resourcehandler.h"

#include "baseobject.h"

Game::Game (const char* levelfile)
	:
	edge (0.0, 0.0),
	collisiondetector (&player, 0.4),
	worldmodifier (this),
	renderer (player.getShape ()),
	hud ((ResourceHandler::instance ()->getDataPath () + "data/hud.xml").c_str ())
{
	score = 0;
	timer = 0.0;
	zoom = 0.0;
	renderer.setZoom (2.0);

	loadLevel (levelfile);
	gamestate = START;
	loop ();
}

Game::~Game ()
{
	Tilemap::iterator row;
	FreeObjects::iterator object;

	for (row = tilemap.begin (); row != tilemap.end (); row++) {
		for (object = row->begin(); object != row->end (); object++) {
			delete (*object).second;
		}
	}
}

void Game::loop ()
{
	int diff = 0;
	int ticks = 0;

	while (!inputcontroller.quit)
	{
		switch (gamestate)
		{
			case START:
				inputcontroller.readInput ();

				if ((diff = (SDL_GetTicks () - ticks)) > 10)
				{
					ticks = SDL_GetTicks ();
					delta_t = diff / 1000.0f;

					animationsystem.update (delta_t);
				
					renderer.drawBackdrop ();
					renderer.render (tilemap, 0.4);
					renderer.render (freeobjects);
					hud.draw ();

					if (inputcontroller.jump || inputcontroller.left || inputcontroller.right) {
						gamestate = MAIN_LOOP;
						ticks = SDL_GetTicks ();
					}

					SDL_Delay (5);
					SDL_GL_SwapBuffers ();
					glClear (GL_COLOR_BUFFER_BIT);
				}
				else {
					SDL_Delay (10 -diff);
				}
			break;

			case MAIN_LOOP:

			inputcontroller.readInput ();

			if ((diff = (SDL_GetTicks () - ticks)) > 10)
			{
				ticks = SDL_GetTicks ();
				delta_t = diff / 1000.0f;
				timer += delta_t;

				if (inputcontroller.left) {
					player.move_left (delta_t);
				}

				if (inputcontroller.right) {
					player.move_right (delta_t);
				}

				if (inputcontroller.jump) {
					player.jump (delta_t);
				}
				else {
					player.reloadPower ();
				}

				update ();
				updateHud ();
				adjustCamera ();

				collisiondetector.runAgainst (tilemap, worldmodifier, delta_t);
				collisiondetector.runAgainst (freeobjects, worldmodifier, delta_t);

				if (player.isGoalReached ()) {
					score += 10000;
					gamestate = END_SEQUENCE;
				}

				float y;
				renderer.getViewport (NULL, &y, NULL);
				if (y - player.getShape ()->gety () > 2.5) {
					gamestate = GAME_OVER;
				}

				edge.setPosition (0.0, y -2.5);

				renderer.drawBackdrop ();
				renderer.render (tilemap, 0.4);
				renderer.render (freeobjects);
				hud.draw ();
				
				SDL_GL_SwapBuffers ();
				glClear (GL_COLOR_BUFFER_BIT);
			}
			else {
				SDL_Delay (10 -diff);
			}
			break;

			case END_SEQUENCE:

				if ((diff = (SDL_GetTicks () - ticks)) > 10)
				{
					ticks = SDL_GetTicks ();
					delta_t = diff / 1000.0f;
					timer += delta_t;

					zoom -= 8.08 * delta_t;
					renderer.setZoom (zoom + 2.0);

					renderer.drawBackdrop ();
					renderer.render (tilemap, 0.4);
					renderer.render (freeobjects);
					hud.draw ();

					SDL_GL_SwapBuffers ();
					glClear (GL_COLOR_BUFFER_BIT);
				}
				else {
					SDL_Delay (10 -diff);
				}

				if (zoom < -1.2)
					gamestate = SHOW_SCORE;
			break;

			case SHOW_SCORE:
				char str[20];
				sprintf (str, "%d", score + player.getNumPowerUps () * 500);
				Hud::sendMessage ("score-points", str);

				sprintf (str, "%d", int(100000.0/timer));
				Hud::sendMessage ("score-bonus", str);

//				cout << "score: " << player.getNumPowerUps () * 500 + score << endl;;
//				cout << "time bonus: " << int(100000.0/timer) << endl;
//				cout << "total score: " << player.getNumPowerUps () * 500 + int(100000.0/timer) + score << endl;

				score = player.getNumPowerUps () * 500 + int(100000.0/timer) + score;

				sprintf (str, "%d", score);
				Hud::sendMessage ("score-total", str);

				gamestate = WAIT_UNTIL_EXIT;

				Hud::turnOn ("score-board");
				Hud::turnOn ("score-total-text");
				Hud::turnOn ("score-total");
				Hud::turnOn ("score-points-text");
				Hud::turnOn ("score-points");
				Hud::turnOn ("score-bonus-text");
				Hud::turnOn ("score-bonus");
				Hud::turnOff ("power");
				Hud::turnOff ("power-bg");
				Hud::turnOff ("power-frame");
			break;

			case GAME_OVER:
				Hud::turnOn ("gameover");

				gamestate=WAIT_UNTIL_EXIT;
			break;

			case WAIT_UNTIL_EXIT:
				inputcontroller.readInput ();

				renderer.drawBackdrop ();
				renderer.render (tilemap, 0.4);
				renderer.render (freeobjects);
				hud.draw ();

				SDL_Delay (5);
				SDL_GL_SwapBuffers ();
				glClear (GL_COLOR_BUFFER_BIT);
			break;
		}
	}
}

void Game::update ()
{
	Tilemap::iterator row;
	FreeObjects::iterator object;

	for (row = tilemap.begin (); row != tilemap.end (); row++) {
		for (object = row->begin(); object != row->end (); object++) {
			(*object).second->update (delta_t, worldmodifier);
		}
	}

	for (object = freeobjects.begin (); object != freeobjects.end (); object++) {
		(*object).second->update (delta_t, worldmodifier);
	}

	animationsystem.update (delta_t);
}

void Game::updateHud ()
{
	char str[20];

	sprintf (str, "%.1f", timer);
	hud.sendMessage ("timer", str);

	sprintf (str, "%10d", score + player.getNumPowerUps () * 500);
	//hud.setTextField ("score", str);
}

void Game::adjustCamera ()
{
	cam_prev = cam_curr;
	cam_curr = player.getPlayerShape ().gety ();

	float cam_diff = cam_curr - cam_prev;

	if (cam_diff > 0.01)
		zoom += 2.2 * 1.0/(1.0 + zoom) * delta_t;
	else
		zoom -= 1.5 * delta_t;

	if (zoom < 0.0)
		zoom = 0.0;

	if (zoom > 2.0)
		zoom = 2.0;

	/*
	float speed = player.getPlayerShape ().vertical_velocity;

	if (speed > 1.0)
		zoom += 2.4 * delta_t;
	else
		zoom -= 1.5 * delta_t;

	if (zoom < 0.0)
		zoom = 0.0;

	if (zoom > 4.0)
		zoom = 4.0;

	*/

	renderer.setZoom (zoom + 2.0);
}

void Game::loadLevel (const char* levelfile)
{
	TiXmlDocument document(levelfile);
	TiXmlHandle handle (&document);
	
	if (!document.LoadFile ())
	{
		document.Clear ();
		throw domain_error ("couldn't open " + string (levelfile));
	}

	/** LOAD Settings **/
	TiXmlElement* player_node = handle.FirstChild("superbungurulevel").FirstChild("settings").FirstChild("player").Element ();

	PlayerShape& playershape = player.getPlayerShape ();
	player_node->QueryFloatAttribute ("startpos_x", &playershape.x);
	player_node->QueryFloatAttribute ("startpos_y", &playershape.y);	

	/** LOAD tilemap **/
	TiXmlElement* row_node = handle.FirstChild("superbungurulevel").FirstChild("tilemap").FirstChild ("row").Element ();
	TiXmlElement* object_node = row_node->FirstChildElement();

	for (; row_node; row_node = row_node->NextSiblingElement ()) {
		FreeObjects row;

		for (object_node = row_node->FirstChildElement(); object_node; object_node = object_node->NextSiblingElement ()) {
			worldmodifier.insertTileObject (objectfactory.createObject (object_node->Value (), object_node), row);
		}
		
		tilemap.push_back (row);
	}
	
	/** LOAD free objects **/
	object_node = handle.FirstChild("superbungurulevel").FirstChild("freeobjects").Element ();
	
	for (object_node = object_node->FirstChildElement (); object_node; object_node = object_node->NextSiblingElement ()) {
		worldmodifier.insertFreeObject (objectfactory.createObject (object_node->Value (), object_node));
	}

	worldmodifier.insertFreeObject (&player); // NOTE temporary
	worldmodifier.insertFreeObject (&edge);
}
