#include <GL/gl.h>
#include <GL/glu.h>
#include <iostream>
#include <stdexcept>
#include <SDL/SDL.h>

#include "hud.h"
#include "widgetfactory.h"
#include "../tinyxml/tinyxml.h"

using std::domain_error;

using namespace SimpleHud;

vector< map<string, Widget*> > Hud::layers;

Hud::Hud (const char* hud_scheme)
{
	TiXmlDocument document(hud_scheme);
	TiXmlHandle handle (&document);

	if (!document.LoadFile ())  {
		throw domain_error ("Hud: Failed to open " + string (hud_scheme));
	}

	TiXmlElement* hud_node = handle.FirstChild ("hud-scheme").Element ();

	if (hud_node->QueryFloatAttribute("width", &width) != TIXML_SUCCESS)
		throw domain_error ("Hud: missing width attribute");

	if (hud_node->QueryFloatAttribute("height", &height) != TIXML_SUCCESS)
		throw domain_error ("Hud: missing height attribute");

	hud_node = hud_node->FirstChildElement ();

	while (hud_node) {
		WidgetFactory::create (hud_node);
		hud_node = hud_node->NextSiblingElement ();
	}
}

Hud::~Hud ()
{
	vector< map<string, Widget*> >::iterator layer;

	for (layer = layers.begin (); layer != layers.end (); layer++) {
		map<string, Widget*>::iterator widget;

		for (widget = layer->begin (); widget != layer->end (); widget++) {
			delete widget->second;
		}

		layer->clear ();
	}

	layers.clear ();
}

void Hud::draw ()
{
	set2dMode ();
	
	vector< map<string, Widget*> >::iterator layer;

	for (layer = layers.begin (); layer != layers.end (); layer++) {
		map<string, Widget*>::iterator widget;

		for (widget = layer->begin (); widget != layer->end (); widget++) {
			if (widget->second->getStatus ())
				widget->second->draw ();
		}
	}

	unset2dMode ();
}
	
void Hud::registerWidget (Widget* widget, string id, int layer)
{
	while ((int)layers.size () <= layer) {
		Layer new_layer;
		layers.push_back (new_layer);
	}

	(layers.at (layer))[id] = widget;
}

void Hud::sendMessage (string id, string message)
{
	vector< map<string, Widget*> >::iterator layer;

	for (layer = layers.begin (); layer != layers.end (); layer++) {
		map<string, Widget*>::iterator widget = layer->find (id);

		if (widget != layer->end ()) {
			widget->second->handleMessage (message);
		}
	}
}

void Hud::sendMessage (string id, float message)
{
	vector< map<string, Widget*> >::iterator layer;

	for (layer = layers.begin (); layer != layers.end (); layer++) {
		map<string, Widget*>::iterator widget = layer->find (id);

		if (widget != layer->end ()) {
			widget->second->handleMessage (message);
		}
	}
}

void Hud::turnOff (string id)
{
	vector< map<string, Widget*> >::iterator layer;

	for (layer = layers.begin (); layer != layers.end (); layer++) {
		map<string, Widget*>::iterator widget = layer->find (id);

		if (widget != layer->end ()) {
			widget->second->setStatus (false);
		}
	}
}

void Hud::turnOn (string id)
{
	vector< map<string, Widget*> >::iterator layer;

	for (layer = layers.begin (); layer != layers.end (); layer++) {
		map<string, Widget*>::iterator widget = layer->find (id);

		if (widget != layer->end ()) {
			widget->second->setStatus (true);
		}
	}
}

void Hud::set2dMode ()
{
	SDL_Surface *screen = SDL_GetVideoSurface();

   glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_DEPTH_TEST);
   glDisable(GL_CULL_FACE);
   glEnable(GL_TEXTURE_2D);

   glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

   glViewport(0, 0, screen->w, screen->h);

   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();

	glOrtho(0.0, width, height, 0.0, 0.0, 1.0);

   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glLoadIdentity();

   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
}

void Hud::unset2dMode ()
{
	glMatrixMode(GL_MODELVIEW);
   glPopMatrix();

   glMatrixMode(GL_PROJECTION);
   glPopMatrix();

   glPopAttrib();

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable (GL_BLEND);
   
   glMatrixMode (GL_MODELVIEW);
}
