#include <GL/gl.h>
#include <cmath>
#include <iostream>
#include <stdexcept>

#include "leveleditor.h"
#include "rock.h"
#include "tinyxml/tinyxml.h"
#include "resourcehandler.h"

using std::domain_error;

LevelEditor::LevelEditor (const char* levelfile)
	:
	cursor (0.0, 0.0),
	editmode (tilemode),
	worldmodifier (this),
	hud ((ResourceHandler::instance ()->getDataPath () + "data/editor_hud.xml").c_str ())
{
	cursor_x = 0.0;
	cursor_y = 0.0;
	select = 0;

	viewport_x = 0.0;
	viewport_y = 0.0;

	zoom = 2.0;

	settings.startposition_x = 0.0;
	settings.startposition_y = 0.0;

	objectlist.push_back ("rock");
	objectlist.push_back ("easyrock");
	objectlist.push_back ("normalrock");
	objectlist.push_back ("hardrock");
	objectlist.push_back ("goal");
	objectlist.push_back ("powerup");
	objectlist.push_back ("playerstart");

	worldmodifier.insertFreeObject (&cursor);

	SDL_WM_GrabInput (SDL_GRAB_ON);
	SDL_WarpMouse (50, 50);
	SDL_Event event;
	while (SDL_PollEvent (&event) != 0);

	load (levelfile);
	edit ();
	save (levelfile);
}

LevelEditor::~LevelEditor ()
{
	/* FIXME clean up */

	SDL_WM_GrabInput (SDL_GRAB_OFF);
}

void LevelEditor::edit ()
{
	while (!inputcontroller.quit)
	{
			inputcontroller.readInput ();

			if (inputcontroller.mouse_moved) {
				cursor_x += inputcontroller.motion_x/50.0;
				cursor_y -= inputcontroller.motion_y/50.0;
 				cursor.setPosition (cursor_x, cursor_y);
			}

			if (inputcontroller.rshift) {
				if (inputcontroller.wheel_up) {
					switchObjectUp ();
					inputcontroller.wheel_up = false;
				}

				if (inputcontroller.wheel_down) {
					switchObjectDown ();
					inputcontroller.wheel_down = false;
				}
			}

			if (inputcontroller.digit(1)) {
				editmode = tilemode;
				hud.sendMessage ("mode", "Tile Mode");
			}
		
			if (inputcontroller.digit(2)) {
				editmode = freemode;
				hud.sendMessage ("mode", "Free Mode");
			}

			if (inputcontroller.digit (3)) {
				editmode = settingsmode;
				hud.sendMessage ("mode", "Select Start Location");
			}

			if (inputcontroller.wheel_up)
				zoom += 0.1;

			if (inputcontroller.wheel_down)
				zoom -= 0.1;

			if (inputcontroller.left_click)
				placeObject (cursor_x, cursor_y);

			if (inputcontroller.right_click)
				removeObject (snapToGrid(cursor_x), snapToGrid(cursor_y));

			if (viewport_y - cursor_y < -zoom)
				viewport_y += 0.1;

			if (viewport_y - cursor_y > zoom)
				viewport_y -= 0.1;

			renderer.setViewport (viewport_x, viewport_y, zoom);
			renderer.render (tilemap, 0.4);
			renderer.render (freeobjects);
			hud.draw ();

			SDL_Delay (5);
			SDL_GL_SwapBuffers ();
			glClear (GL_COLOR_BUFFER_BIT);
	}
}

void LevelEditor::save (const char* levelfile)
{
	string levelname (levelfile);
	levelname = levelname.substr (levelname.find_last_of ('/') +1);

	TiXmlDocument document (levelfile);
	TiXmlElement body_node ("superbungurulevel");

	body_node.SetAttribute ("name", levelname.c_str ());

	TiXmlElement settings_node ("settings");

	TiXmlElement player_node ("player");
	player_node.SetDoubleAttribute ("startpos_x", settings.startposition_x);
	player_node.SetDoubleAttribute ("startpos_y", settings.startposition_y);
	settings_node.InsertEndChild (player_node);

	body_node.InsertEndChild (settings_node);


	TiXmlElement tilemap_node ("tilemap");
	Tilemap::iterator row;
	FreeObjects::iterator object;

	for (row = tilemap.begin (); row != tilemap.end (); row++) {
		TiXmlElement row_node ("row");
		for (object = row->begin(); object != row->end(); object++) {
			row_node.InsertEndChild (object->second->getXMLRepresentation ());
		}
		tilemap_node.InsertEndChild (row_node);
	}

	body_node.InsertEndChild (tilemap_node);

	TiXmlElement freeobjects_node ("freeobjects");
	for (object = ++freeobjects.begin (); object != freeobjects.end (); object++) {
		freeobjects_node.InsertEndChild (object->second->getXMLRepresentation ());
	}

	body_node.InsertEndChild (freeobjects_node);
	
	document.InsertEndChild (body_node);
	document.SaveFile ();
}

void LevelEditor::load (const char* levelfile)
{
	TiXmlDocument document(levelfile);
	TiXmlHandle handle (&document);
	
	if (!document.LoadFile ())
	{
		if (strlen (levelfile) == 0)
			throw domain_error ("LevelEditor: Can't create a level without a name");

		loadTemplate ();
		return;
	}

	TiXmlElement* player_node = handle.FirstChild("superbungurulevel").FirstChild("settings").FirstChild("player").Element ();

	player_node->QueryFloatAttribute ("startpos_x", &settings.startposition_x);
	player_node->QueryFloatAttribute ("startpos_y", &settings.startposition_y);	

	TiXmlElement* row_node = handle.FirstChild("superbungurulevel").FirstChild("tilemap").FirstChild ("row").Element ();
	TiXmlElement* object_node = row_node->FirstChildElement();

	for (; row_node; row_node = row_node->NextSiblingElement ()) {
		FreeObjects row;

		for (object_node = row_node->FirstChildElement(); object_node; object_node = object_node->NextSiblingElement ()) {
			worldmodifier.insertTileObject (objectfactory.createObject (object_node->Value (), object_node), row);
		}
		
		tilemap.push_back (row);
	}

	object_node = handle.FirstChild("superbungurulevel").FirstChild("freeobjects").Element ();
	
	for (object_node = object_node->FirstChildElement (); object_node; object_node = object_node->NextSiblingElement ()) {
		worldmodifier.insertFreeObject (objectfactory.createObject (object_node->Value (), object_node));
	}
}

void LevelEditor::loadTemplate ()
{
	for (float y = 0.0; y <= 10.0; y += 0.40) {
		FreeObjects row;
		for (float x = -2.0; x <= 2.0; x += 0.40) {
			if (x <= -1.60 || x >= 1.40 || y < 0.80)
				worldmodifier.insertTileObject (new Rock (x,y, indestructible), row);
		}
		tilemap.push_back (row);
	}
}

void LevelEditor::switchObjectUp ()
{
	select++;

	if (select > 6)
		select = 0;

	string path ("data/gfx/");
	path = path + objectlist[select] + ".png";

	hud.sendMessage ("selector", path.c_str ());
}
void LevelEditor::switchObjectDown ()
{
	select--;

	if (select < 0)
		select = 6;

	string path ("data/gfx/");
	path = path + objectlist[select] + ".png";

	hud.sendMessage ("selector", path.c_str ());
}

void LevelEditor::placeObject (float x, float y)
{
	switch (editmode) 
	{
		case tilemode: 
		{
			float _x = snapToGrid (x);
			float _y = snapToGrid (y);

			float at_row = y / 0.4;

			if (tilemap.size ()-1 < at_row)
				addRows ((int)(at_row - (tilemap.size ()-1)));

			Tilemap::iterator row = tilemap.begin () + (int)at_row;
			worldmodifier.insertTileObject (objectfactory.createObject (objectlist[select], _x, _y), *row); 
		}
		break;

		case freemode:
			worldmodifier.insertFreeObject (objectfactory.createObject (objectlist[select], x, y));
		break;

		case settingsmode:
			settings.startposition_x = x;
			settings.startposition_y = y;
		break;
	}
}

void LevelEditor::removeObject (float x, float y)
{
	Tilemap::iterator row;
	FreeObjects::iterator object;

	switch (editmode) {
		case tilemode:
			for (row = tilemap.begin (); row != tilemap.end (); row++) {
				for (object = row->begin(); object != row->end (); object++) {
					if ((*object).second->getShape ()->check (cursor.getShape ()))
						worldmodifier.deleteObject ((*object).second->getSignature ());
				}
			}
		break;

		case freemode:
			for (object = ++freeobjects.begin(); object != freeobjects.end (); object++) {
				if ((*object).second->getShape ()->check (cursor.getShape ()))
					worldmodifier.deleteObject ((*object).second->getSignature ());
			}
		break;

		default:
			// nothing
		break;
	}

	worldmodifier.execute ();
}

void LevelEditor::addRows (int num_rows)
{
	for (int i = 0; i < num_rows; i++) {
		FreeObjects new_row;
		tilemap.push_back (new_row);
	}
}

float LevelEditor::snapToGrid (float pos)
{
	float snap_distance = pos - (int)(pos/0.4)*0.4;

	if (snap_distance > 0) {
		if (snap_distance > 0.4/2.0)
			return ((int)(pos/0.4)*0.4) + 0.4;
		else
			return ((int)(pos/0.4)*0.4);
	}
	else {
		if (-snap_distance > 0.4/2.0)
			return ((int)(pos/0.4)*0.4) - 0.4;
		else
			return ((int)(pos/0.4)*0.4);
	}
}
