#include <iostream>

#include "menu/menusystem.h"
#include "menu/callback.h"
#include "resourcehandler.h"
#include "openglwindow.h"
#include "leveleditor.h"
#include "renderutil.h"
#include "highscores.h"
#include "progress.h"
#include "game.h"

using namespace std;

class QuitCallback : public Callback
{
	public:

	QuitCallback (MenuSystem& menusystem)
	: menusystem (menusystem) { }

	void execute ()
	{
		menusystem.shutdown ();
	}

	private:

	MenuSystem& menusystem;
};

class NameInputCallback : public Callback
{
	public:

	NameInputCallback (MenuSystem& menusystem, HighScores& highscores)
	: highscores (highscores),
	  menusystem (menusystem) { }

	void execute ()
	{
		ResourceHandler* resource_handler = ResourceHandler::instance ();

		highscores.updateRecord (score, levelname, string_arg1);
		highscores.saveHighScores ();
		menusystem.popMenuHistory ();
		menusystem.loadMenu ((resource_handler->getDataPath () + "data/menus/highscoremenu.xml").c_str ());
		menusystem.showMenu ("highscoremenu");
	}

	void setLevelName (string levelname)
	{
		this->levelname = levelname;
	}

	void setScore (int score)
	{
		this->score = score;
	}

	private:

	HighScores& highscores;
	MenuSystem& menusystem;
	
	string levelname;
	int score;
};

class PlayCallback : public Callback
{
	public:

	PlayCallback (MenuSystem& menusystem, HighScores& highscores, Progress& progress, NameInputCallback& nameinput_callback)
		: menusystem (menusystem),
		  highscores (highscores),
		  progress (progress),
		  nameinput_callback (nameinput_callback) { }

	void execute ()
	{
		ResourceHandler* resource_handler = ResourceHandler::instance ();

		try {
			Game game ((resource_handler->getDataPath () + string_arg2).c_str ());

			if (game.getScore () > 0) {
				if (progress.update (string_arg2)) {
					menusystem.loadMenu ((resource_handler->getDataPath () + "data/menus/levelmenu.xml").c_str ());
					menusystem.setCallback ("levelmenu", "levelview", this);
				}
			}

			if (highscores.checkRecord (game.getScore (), string_arg1)) {
				nameinput_callback.setLevelName (string_arg1);
				nameinput_callback.setScore (game.getScore ());
		
				menusystem.loadMenu ((resource_handler->getDataPath () + "data/menus/nameinputmenu.xml").c_str ());
				menusystem.setCallback ("nameinput", "input", &nameinput_callback);
				menusystem.showMenu ("nameinput");
			}
		}
		catch (exception& e)
		{
			cout << e.what () << endl;
		}
	}

	private:

	MenuSystem& menusystem;
	HighScores& highscores;
	Progress& progress;
	NameInputCallback& nameinput_callback;
};

class EditCallback : public Callback
{
	public:

	EditCallback (MenuSystem& menusystem, PlayCallback& play_callback)
		: menusystem (menusystem),
		  play_callback (play_callback) { }

	void execute ()
	{
		ResourceHandler* resource_handler = ResourceHandler::instance ();

		try {
			if (!string_arg2.empty ())
				LevelEditor editor ((resource_handler->getDataPath () + string_arg2).c_str ());
			else
				LevelEditor editor ((resource_handler->getDataPath () + "data/userlevels/" + string_arg1).c_str ());

			menusystem.loadMenu ((resource_handler->getDataPath () + "data/menus/editmenu.xml").c_str ());
			menusystem.loadMenu ((resource_handler->getDataPath () + "data/menus/userlevelmenu.xml").c_str ());

			menusystem.setCallback ("userlevelmenu", "fileview", &play_callback);
			menusystem.setCallback ("editmenu", "fileview", this);
			menusystem.setCallback ("editmenu", "newlevel", this);
		}
		catch (exception& e)
		{
			cout << e.what () << endl;
		}
	}

	private:

	MenuSystem& menusystem;
	PlayCallback& play_callback;
};

void printUsage ()
{
	cout << "SuperBunguru Pre-Alpha version. Written and Copyrights by Jacob Persson." << endl;
	cout << "THIS SOFTWARE COMES WITH ABSOLUTELY NO WARRANTY! USE AT YOUR OWN RISK! " << endl;
}

int main (int argc, char** argv)
{
	/*
	if (argc != 2 && argc != 3) {
		printUsage ();
		return 0;
	}
	*/

	ResourceHandler* resource_handler = ResourceHandler::instance ();

	OpenGLWindow openglwindow;
	MenuSystem menusystem;
	HighScores highscores (resource_handler->getDataPath () + "data/highscores.xml");
	Progress progress (resource_handler->getDataPath () + "data/progress.xml");

	try {
		highscores.loadHighScores ();
	}
	catch (exception& e)
	{
		cout << e.what () << endl;
	}

	try {
		menusystem.loadMenu ((resource_handler->getDataPath () + "data/menus/mainmenu.xml").c_str ());
		menusystem.loadMenu ((resource_handler->getDataPath () + "data/menus/editmenu.xml").c_str ());
		menusystem.loadMenu ((resource_handler->getDataPath () + "data/menus/levelmenu.xml").c_str ());
		menusystem.loadMenu ((resource_handler->getDataPath () + "data/menus/highscoremenu.xml").c_str ());
		menusystem.loadMenu ((resource_handler->getDataPath () + "data/menus/userlevelmenu.xml").c_str ());
	}
	catch (exception& e) {
		cout << e.what () << endl;
		return -1;
	}

	NameInputCallback nameinput_callback (menusystem, highscores);
	PlayCallback play_callback (menusystem, highscores, progress, nameinput_callback);
	EditCallback edit_callback (menusystem, play_callback);
	QuitCallback quit_callback (menusystem);

	menusystem.setCallback ("userlevelmenu", "fileview", &play_callback);
	menusystem.setCallback ("levelmenu", "levelview", &play_callback);
	menusystem.setCallback ("editmenu", "fileview", &edit_callback);
	menusystem.setCallback ("editmenu", "newlevel", &edit_callback);
	menusystem.setCallback ("mainmenu", "quit", &quit_callback);

	try {
		openglwindow.setResolution (800, 600, 32, SDL_OPENGL);
		openglwindow.setCaption (".~SuperBunguru~.");
	}
	catch (exception& e) {
		cout << e.what () << endl;
		return -1;
	}
	
	menusystem.showMenu("mainmenu");
	while (!menusystem.quitSignal())
	{
		menusystem.update ();

		SDL_Delay (5);
		SDL_GL_SwapBuffers ();
		glClear (GL_COLOR_BUFFER_BIT);
	}

	try {
		highscores.saveHighScores ();
	}
	catch (exception& e) {
		cout << e.what () << endl;
	}

	RenderUtil::clearCache (); // NOTE move caching to ResourceHandler

	/*	
	try {
		if (strcmp (argv[1], "-e") == 0)
			LevelEditor leveleditor (argv[2]);
		else
			Game game (argv[1]);
	}
	catch (exception& e)
	{
		cout << e.what () << endl;
	}
	*/
}
