#include <GL/gl.h>
#include <GL/glu.h>
#include <string>
#include <iostream>
#include <stdexcept>
#include <SDL/SDL.h>

#include "menusystem.h"
#include "../tinyxml/tinyxml.h"

list<Menu*> MenuSystem::menus;
stack<Menu*> MenuSystem::menu_history;

MenuSystem::MenuSystem ()
{
	SDL_EnableUNICODE(1);

	quit = false;
}

MenuSystem::~MenuSystem ()
{
	list<Menu*>::iterator menu;

	for (menu = menus.begin (); menu != menus.end (); menu++) {
		if ((*menu)->root != 0)
			delete (*menu)->root;
	}
}

bool MenuSystem::quitSignal ()
{
	return quit;
}

void MenuSystem::shutdown ()
{
	quit = true;
}

void MenuSystem::update ()
{
	set2dMode ();
	menu_history.top ()->root->draw (400.0, 300.0, 200.0, 150.0);
	unset2dMode ();

	checkForInput ();
}

void MenuSystem::loadMenu (const char* scheme)
{
	TiXmlDocument document(scheme);
	TiXmlHandle handle (&document);

	if (!document.LoadFile ())  {
		throw domain_error ("MenySystem: Failed to open " + string (scheme));
	}

	TiXmlElement* menu_node = handle.FirstChild ("menu-scheme").Element ();
	string menu_id = menu_node->Attribute("id");

	list<Menu*>::iterator menu;

	for (menu = menus.begin (); menu != menus.end (); menu++) {
		if (menu_id.compare ((*menu)->id) == 0) {
			if ((*menu)->root != 0)
				delete (*menu)->root;

			menus.erase (menu);
			break;
		}
	}

	Menu* new_menu = new Menu;
	menus.push_back(new_menu);

	new_menu->id = menu_id;
	new_menu->root = NULL;
	new_menu->root = new Root (menu_node);
}

void MenuSystem::showMenu (string id)
{
	list<Menu*>::iterator menu;

	for (menu = menus.begin (); menu != menus.end (); menu++) {
		if (id.compare ((*menu)->id) == 0)
			menu_history.push (*menu);
	}
}

void MenuSystem::popMenuHistory ()
{
	menu_history.pop ();
}

void MenuSystem::setCallback (string menu_id, string widget_id, Callback* callback)
{
	for (list<Menu*>::iterator it = menus.begin (); it != menus.end (); it++) {
		if (menu_id.compare ((*it)->id) == 0) {
			for (list<Widget*>::iterator widget = (*it)->widgets.begin (); widget != (*it)->widgets.end (); widget++) {
				if (widget_id.compare ((*widget)->getId ()) == 0 ) {
					(*widget)->setCallback (callback);
				}
			}
		}
	}
}

void MenuSystem::registerWidget (Widget* widget)
{
	menus.back ()->widgets.push_back (widget);
}

void MenuSystem::checkForInput ()
{
	while (SDL_PollEvent (&event)){
		switch (event.type){
			case SDL_QUIT:
				quit = true;
			break;

			case SDL_KEYDOWN:
				if (event.key.keysym.sym == SDLK_ESCAPE) {
					menu_history.pop ();

					if (menu_history.empty ()) {
						quit = true;
						return;
					}
				}
				menu_history.top ()->root->handleKeyPress (event.key.keysym);
			break;
		}
	}
}

void MenuSystem::set2dMode ()
{
	SDL_Surface *screen = SDL_GetVideoSurface();

   glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_DEPTH_TEST);
   glDisable(GL_CULL_FACE);
   glEnable(GL_TEXTURE_2D);

   glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

   glViewport(0, 0, screen->w, screen->h);

   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();

   //glOrtho(0.0, (GLdouble)screen->w, (GLdouble)screen->h, 0.0, 0.0, 1.0);
	glOrtho(0.0, 400.0, 300.0, 0.0, 0.0, 1.0);

   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glLoadIdentity();

   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
}

void MenuSystem::unset2dMode ()
{
	glMatrixMode(GL_MODELVIEW);
   glPopMatrix();

   glMatrixMode(GL_PROJECTION);
   glPopMatrix();

   glPopAttrib();

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable (GL_BLEND);
   
   glMatrixMode (GL_MODELVIEW);
}
