#include <iostream>

#include "hud/hud.h"
#include "player.h"
#include "animation.h"
#include "basicanimations.h"
#include "resourcehandler.h"

#define MAX_COMBO_GAP 0.3
#define POWERJUMP_COMBO 5
#define MAX_FALLSPEED -7.0
#define VELOCITY_CHANGERATE 10.0
#define HORIZONTAL_VELOCITY 2.0
#define JUMP_VELOCITY 6.0
#define POWERJUMP_VELOCITY 12.0
#define RECHARGE_RATE 0.2
#define MAX_JUMP_POWER 0.2
#define NUM_AIR_JUMPS 3
#define MAX_CHARGE_POWER (MAX_JUMP_POWER / NUM_AIR_JUMPS)

using namespace std;

Player::Player ()
	:
	BaseObject (&playershape),
	modifier (mod_normal),
	drawstate (draw_standing),
	playeranimation ("data/gfx/player_sets.png", 2, 2, 0.2),
	beam_animation ("data/gfx/beam.png"),
	scale (0.30, 0.30, -1.0, 0.0)
{
	combo_time = 0.0;
	collected_powerups = 0;
	collected_powerups_combo = 0;

	jump_power = 0.0;
	total_jump_power = MAX_JUMP_POWER;
	reloadPower ();

	goal_reached = false;
	playershape.x = 0.2;
	playershape.y = 2.0;
	playershape.sidelength = 0.15;
	playershape.vertical_velocity = 0.0;

	// animation settings
	playeranimation.addAnimation (&translate);
	playeranimation.addAnimation (&scale);
	playeranimation.addAnimation (&rotate);
	AnimationSystem::newAnimationObject (&playeranimation);

	beam_animation.addAnimation (&translate);
	beam_animation.addAnimation (&scale);
	AnimationSystem::newAnimationObject (&beam_animation);
}

void Player::draw ()
{
	switch (drawstate)
	{
		case draw_standing:
				scale.set (0.25, 0.25);
				translate.set (playershape.x, playershape.y);
				playeranimation.setRate (0.20);
				playeranimation.selectSet (0);
				playeranimation.animate ();
		break;

		case draw_falling:
			scale.set (0.25, 0.25);
			translate.set (playershape.x, playershape.y + 0.10);
			playeranimation.setRate (0);
			playeranimation.selectSet (1);
			playeranimation.selectFrame (1);
			playeranimation.animate ();
		break;

		case draw_jump:
				scale.set (0.25, 0.25);
				translate.set (playershape.x, playershape.y);
				playeranimation.setRate (0);
				playeranimation.selectSet (1);
				playeranimation.selectFrame (0);
				playeranimation.animate ();
		break;

		case draw_powerjump:
				scale.set (0.20, 0.20 * (1 + (playershape.y - powerjump_start)/0.4));
				translate.set (playershape.x, (playershape.y -powerjump_start)*0.5 + powerjump_start);
				beam_animation.animate ();
		break;
	}
}

TiXmlNode& Player::getXMLRepresentation ()
{
	TiXmlElement player_node ("player");

	return player_node;
}


void Player::reloadPower ()
{
	float needed_power = (MAX_JUMP_POWER / NUM_AIR_JUMPS) - jump_power;

	if (total_jump_power > needed_power) {
		total_jump_power -= needed_power;
		jump_power += needed_power;
	}
	else {
		jump_power += total_jump_power;
		total_jump_power = 0.0;
	}
}

void Player::rechargePower (float delta_t)
{
	total_jump_power += delta_t * RECHARGE_RATE;

	collected_powerups_combo = 0;
	combo_time = 0.0;
	modifier = mod_normal;
	drawstate = draw_standing;

	if ((total_jump_power  + jump_power) > MAX_JUMP_POWER) {
		total_jump_power = MAX_JUMP_POWER - jump_power;
	}

	reloadPower ();


	// TEMPORARY
	rotate.reset ();
	rotate.set (0.0);
}

void Player::collectPowerUp ()
{
	++collected_powerups;
	++collected_powerups_combo;
	combo_time = 0.0;

	total_jump_power += MAX_JUMP_POWER / 5.0;

	if (total_jump_power + jump_power > MAX_JUMP_POWER)
		total_jump_power = MAX_JUMP_POWER - jump_power;

	if (collected_powerups_combo >= POWERJUMP_COMBO) {
		modifier = mod_powerjump;
		jump_power = 0.0;
		total_jump_power = MAX_JUMP_POWER;
		collected_powerups_combo = 0;
		combo_time  = 0.0;

		reloadPower ();
	}

	//reloadPower ();
}

void Player::goalReached ()
{
	goal_reached = true;
}

bool Player::isGoalReached ()
{
	return goal_reached;
}

void Player::move_left (float delta_t)
{
	playershape.x -= HORIZONTAL_VELOCITY * delta_t;
	
	rotate.set (400.0);
}

void Player::move_right (float delta_t)
{
	playershape.x += HORIZONTAL_VELOCITY * delta_t;

	rotate.set (-400.0);
}

void Player::jump (float delta_t)
{
	if (modifier == mod_normal) {
		if (jump_power > 0.0) {
			playershape.vertical_velocity = JUMP_VELOCITY * jump_power / MAX_CHARGE_POWER;
			//jump_power -= delta_t;
			jump_power = 0.0;
			drawstate = draw_jump;
		}

		if (jump_power < 0.0)
		{
			jump_power = 0.0;
		}
	} 
	else if (modifier == mod_powerjump) {
		if (jump_power > 0.0) {
			playershape.vertical_velocity = POWERJUMP_VELOCITY * jump_power / MAX_CHARGE_POWER;
			jump_power = 0.0;
			drawstate = draw_powerjump;
			powerjump_start = playershape.y;
		}

		if (jump_power <= 0.0) {
			jump_power = 0.0;
			modifier = mod_normal;
		}
	}
}

void Player::update (float delta_t, WorldModifier& worldmodifier)
{
	playershape.vertical_velocity -= VELOCITY_CHANGERATE * delta_t;

	if (playershape.vertical_velocity < MAX_FALLSPEED)
		playershape.vertical_velocity = MAX_FALLSPEED;

	playershape.y += playershape.vertical_velocity * delta_t;

	combo_time += delta_t;

	if (combo_time > MAX_COMBO_GAP)
		collected_powerups_combo = 0;

	if (drawstate == draw_powerjump && playershape.vertical_velocity <= 0.0)
		drawstate = draw_standing;

	if (drawstate == draw_jump && playershape.vertical_velocity <= 0.0)
		drawstate = draw_falling;

	Hud::sendMessage ("power", (total_jump_power + jump_power)/ MAX_JUMP_POWER);
}
